#!/usr/bin/env python

from subprocess import Popen, PIPE
from sympy import *
from MVS import Matrice, VectorSpace

class GAPInterface:
    def __init__(self):
        self.process = Popen(["gap", "-q"], stdin = PIPE, stdout = PIPE, bufsize = 1, close_fds = True)
        self.input = self.process.stdin
        self.output = self.process.stdout

    def write(self, s, debug=False):
        if debug: print(s)
        self.input.write(s + "\n")

    def read(self):
        s = self.output.readline()
        while s[-2] == " ":
            s = s[:-1]
            s += self.output.readline()
        return s

    def writeread(self, s, debug=False):
        self.write(s, debug)
        return self.read()

    def __del__(self):
        self.write("quit;")

    def rewrite_expr(self, e):
        if type(e) == type(Matrice(zero(1))):
            R = e.lines
            C = e.cols
            n = e.tolist()
            s = '[' 
            for r in xrange(R):
                s += '['
                for c in xrange(C):
                    s += self.rewrite_expr(n[r][c]) + ", "
                s += '], '
            s += ']'
        elif e.is_Number:
            s = e.__str__()
        elif e.is_Mul:
            t = e.as_two_terms()
            s = self.rewrite_expr(t[0]) + "*" + self.rewrite_expr(t[1])
        elif e.func == exp:
            p = e.args[0].as_coefficient(pi*I).p
            q = e.args[0].as_coefficient(pi*I).q
            if p % 2 == 0:
                s = "E(%d)^%d" % (q, (p/2) % q)
            else:
                s = "E(%d)^%d" % (2*q, p % (2*q))
        elif e == I:
            s = "E(4)"
        else:
            raise("Do not know how to convert " + e.__str__())
        return s

    def NullspaceMat(self, M):
        s = self.writeread("NullspaceMat(" + self.rewrite_expr(M) + ");").replace("^", "**")
        def E(n):
            return exp(2*pi*I/n)
        exec("r = " + s) in locals()
        ret = []
        for x in r:
            ret.append(Matrice(x))
        return ret
            
    def IdSmallGroup(self, l, pg, ag):
        """
        Returns the id and the structure description of the group generated by
        l, quotiented by the groups generated by pg and ag.
        
        l: list of matrices that generates the group
        
        pg: the generators to quotient to obtain a projective matrix group (may
            give a group which is not contained in <l> - in this case, there's
            no need to quotient)
              
        ag: the generator of the action - should always be contained in <l>
        """
        s = "G := Group("
        for x in l:
            s += self.rewrite_expr(x) + ","
        s = s[:-1] + ");;"
        self.write(s)

        self.write("H := Group(" + self.rewrite_expr(pg) + ");;")
        self.write("x := " + self.rewrite_expr(ag) + ";;")
        o = self.writeread("IsSubgroup(G, H);")
        if o == "true\n":
            self.write("p := NaturalHomomorphismByNormalSubgroup(G, H);;")
            self.write("x := ImagesRepresentative(p, x);;")
            self.write("G := Image(p);;")

        self.write("K := Group(x);;")
        o = self.writeread("IsNormal(G, K);")
        if o == "false\n": raise("Action subgroup not contained or not normal.")
        self.write("p := NaturalHomomorphismByNormalSubgroup(G, K);;")
        o = self.writeread("IdSmallGroup(Image(p));")
        exec("r = " + o)
        o = self.writeread("StructureDescription(Image(p));")
        exec("g = " + o[:-1])
        return r[0], r[1], g
